@@ -16,45 +16,52 @@
 // under the License.
 
 use core::fmt::Debug;
-use std::collections::HashMap;
+use std::fmt::Formatter;
+use std::sync::Arc;
 
-use async_trait::async_trait;
+use bytes::Buf;
 use http::header::CONTENT_LENGTH;
 use http::header::CONTENT_TYPE;
 use http::Request;
 use http::Response;
 use http::StatusCode;
 use log::debug;
+use serde::Deserialize;
 use tokio::sync::OnceCell;
 
+use super::delete::WebhdfsDeleter;
 use super::error::parse_error;
-use super::error::parse_error_msg;
 use super::lister::WebhdfsLister;
 use super::message::BooleanResp;
 use super::message::FileStatusType;
 use super::message::FileStatusWrapper;
 use super::writer::WebhdfsWriter;
+use super::writer::WebhdfsWriters;
 use crate::raw::*;
+use crate::services::WebhdfsConfig;
 use crate::*;
 
 const WEBHDFS_DEFAULT_ENDPOINT: &str = "http://127.0.0.1:9870";
 
+impl Configurator for WebhdfsConfig {
+    type Builder = WebhdfsBuilder;
+    fn into_builder(self) -> Self::Builder {
+        WebhdfsBuilder { config: self }
+    }
+}
+
 /// [WebHDFS](https://hadoop.apache.org/docs/stable/hadoop-project-dist/hadoop-hdfs/WebHDFS.html)'s REST API support.
 #[doc = include_str!("docs.md")]
 #[derive(Default, Clone)]
 pub struct WebhdfsBuilder {
-    root: Option<String>,
-    endpoint: Option<String>,
-    delegation: Option<String>,
-    disable_list_batch: bool,
+    config: WebhdfsConfig,
 }
 
 impl Debug for WebhdfsBuilder {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        f.debug_struct("Builder")
-            .field("root", &self.root)
-            .field("endpoint", &self.endpoint)
-            .finish_non_exhaustive()
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        let mut d = f.debug_struct("WebhdfsBuilder");
+        d.field("config", &self.config);
+        d.finish_non_exhaustive()
     }
 }
 
@@ -66,10 +73,12 @@ impl WebhdfsBuilder {
     /// # Note
     ///
     /// The root will be automatically created if not exists.
-    pub fn root(&mut self, root: &str) -> &mut Self {
-        if !root.is_empty() {
-            self.root = Some(root.to_string())
-        }
+    pub fn root(mut self, root: &str) -> Self {
+        self.config.root = if root.is_empty() {
+            None
+        } else {
+            Some(root.to_string())
+        };
 
         self
     }
@@ -84,10 +93,10 @@ impl WebhdfsBuilder {
     ///
     /// If user inputs endpoint without scheme, we will
     /// prepend `http://` to it.
-    pub fn endpoint(&mut self, endpoint: &str) -> &mut Self {
+    pub fn endpoint(mut self, endpoint: &str) -> Self {
         if !endpoint.is_empty() {
             // trim tailing slash so we can accept `http://127.0.0.1:9870/`
-            self.endpoint = Some(endpoint.trim_end_matches('/').to_string());
+            self.config.endpoint = Some(endpoint.trim_end_matches('/').to_string());
         }
         self
     }
@@ -98,9 +107,9 @@ impl WebhdfsBuilder {
     /// # Note
     /// The builder prefers using delegation token over username.
     /// If both are set, delegation token will be used.
-    pub fn delegation(&mut self, delegation: &str) -> &mut Self {
+    pub fn delegation(mut self, delegation: &str) -> Self {
         if !delegation.is_empty() {
-            self.delegation = Some(delegation.to_string());
+            self.config.delegation = Some(delegation.to_string());
         }
         self
     }
@@ -110,29 +119,30 @@ impl WebhdfsBuilder {
     /// # Note
     ///
     /// When listing a directory, the backend will default to use batch listing.
-    /// If disable, the backend will list all files/directories in one request.
-    pub fn disable_list_batch(&mut self) -> &mut Self {
-        self.disable_list_batch = true;
+    /// If disabled, the backend will list all files/directories in one request.
+    pub fn disable_list_batch(mut self) -> Self {
+        self.config.disable_list_batch = true;
+        self
+    }
+
+    /// Set temp dir for atomic write.
+    ///
+    /// # Notes
+    ///
+    /// If not set, write multi not support, eg: `.opendal_tmp/`.
+    pub fn atomic_write_dir(mut self, dir: &str) -> Self {
+        self.config.atomic_write_dir = if dir.is_empty() {
+            None
+        } else {
+            Some(String::from(dir))
+        };
         self
     }
 }
 
 impl Builder for WebhdfsBuilder {
     const SCHEME: Scheme = Scheme::Webhdfs;
-    type Accessor = WebhdfsBackend;
-
-    fn from_map(map: HashMap<String, String>) -> Self {
-        let mut builder = WebhdfsBuilder::default();
-
-        map.get("root").map(|v| builder.root(v));
-        map.get("endpoint").map(|v| builder.endpoint(v));
-        map.get("delegation").map(|v| builder.delegation(v));
-        map.get("disable_list_batch")
-            .filter(|v| v == &"true")
-            .map(|_| builder.disable_list_batch());
-
-        builder
-    }
+    type Config = WebhdfsConfig;
 
     /// build the backend
     ///
@@ -140,15 +150,15 @@ impl Builder for WebhdfsBuilder {
     ///
     /// when building backend, the built backend will check if the root directory
     /// exits.
-    /// if the directory does not exits, the directory will be automatically created
-    fn build(&mut self) -> Result<Self::Accessor> {
+    /// if the directory does not exit, the directory will be automatically created
+    fn build(self) -> Result<impl Access> {
         debug!("start building backend: {:?}", self);
 
-        let root = normalize_root(&self.root.take().unwrap_or_default());
+        let root = normalize_root(&self.config.root.unwrap_or_default());
         debug!("backend use root {root}");
 
         // check scheme
-        let endpoint = match self.endpoint.take() {
+        let endpoint = match self.config.endpoint {
             Some(endpoint) => {
                 if endpoint.starts_with("http") {
                     endpoint
@@ -160,10 +170,9 @@ impl Builder for WebhdfsBuilder {
         };
         debug!("backend use endpoint {}", endpoint);
 
-        let auth = self
-            .delegation
-            .take()
-            .map(|dt| format!("delegation_token={dt}"));
+        let atomic_write_dir = self.config.atomic_write_dir;
+
+        let auth = self.config.delegation.map(|dt| format!("delegation={dt}"));
 
         let client = HttpClient::new()?;
 
@@ -173,7 +182,8 @@ impl Builder for WebhdfsBuilder {
             auth,
             client,
             root_checker: OnceCell::new(),
-            disable_list_batch: self.disable_list_batch,
+            atomic_write_dir,
+            disable_list_batch: self.config.disable_list_batch,
         };
 
         Ok(backend)
@@ -188,59 +198,181 @@ pub struct WebhdfsBackend {
     auth: Option<String>,
     root_checker: OnceCell<()>,
 
+    pub atomic_write_dir: Option<String>,
     pub disable_list_batch: bool,
     pub client: HttpClient,
 }
 
 impl WebhdfsBackend {
-    /// create object or make a directory
-    ///
-    /// TODO: we should split it into mkdir and create
-    pub fn webhdfs_create_object_request(
-        &self,
-        path: &str,
-        size: Option<usize>,
-        args: &OpWrite,
-        body: AsyncBody,
-    ) -> Result<Request<AsyncBody>> {
+    pub fn webhdfs_create_dir_request(&self, path: &str) -> Result<Request<Buffer>> {
         let p = build_abs_path(&self.root, path);
-        let op = if path.ends_with('/') {
-            "MKDIRS"
-        } else {
-            "CREATE"
-        };
+
         let mut url = format!(
-            "{}/webhdfs/v1/{}?op={}&overwrite=true",
+            "{}/webhdfs/v1/{}?op=MKDIRS&overwrite=true&noredirect=true",
             self.endpoint,
             percent_encode_path(&p),
-            op,
         );
         if let Some(auth) = &self.auth {
             url += format!("&{auth}").as_str();
         }
 
-        let mut req = Request::put(&url);
+        let req = Request::put(&url);
 
-        // mkdir does not redirect
-        if path.ends_with('/') {
-            return req.body(AsyncBody::Empty).map_err(new_request_build_error);
+        req.body(Buffer::new()).map_err(new_request_build_error)
+    }
+    /// create object
+    pub async fn webhdfs_create_object_request(
+        &self,
+        path: &str,
+        size: Option<u64>,
+        args: &OpWrite,
+        body: Buffer,
+    ) -> Result<Request<Buffer>> {
+        let p = build_abs_path(&self.root, path);
+
+        let mut url = format!(
+            "{}/webhdfs/v1/{}?op=CREATE&overwrite=true&noredirect=true",
+            self.endpoint,
+            percent_encode_path(&p),
+        );
+        if let Some(auth) = &self.auth {
+            url += format!("&{auth}").as_str();
         }
 
+        let req = Request::put(&url);
+
+        let req = req.body(Buffer::new()).map_err(new_request_build_error)?;
+
+        let resp = self.client.send(req).await?;
+
+        let status = resp.status();
+
+        if status != StatusCode::CREATED && status != StatusCode::OK {
+            return Err(parse_error(resp));
+        }
+
+        let bs = resp.into_body();
+
+        let resp: LocationResponse =
+            serde_json::from_reader(bs.reader()).map_err(new_json_deserialize_error)?;
+
+        let mut req = Request::put(&resp.location);
+
         if let Some(size) = size {
-            req = req.header(CONTENT_LENGTH, size.to_string());
-        }
+            req = req.header(CONTENT_LENGTH, size);
+        };
+
         if let Some(content_type) = args.content_type() {
             req = req.header(CONTENT_TYPE, content_type);
+        };
+        let req = req.body(body).map_err(new_request_build_error)?;
+
+        Ok(req)
+    }
+
+    pub async fn webhdfs_init_append_request(&self, path: &str) -> Result<String> {
+        let p = build_abs_path(&self.root, path);
+        let mut url = format!(
+            "{}/webhdfs/v1/{}?op=APPEND&noredirect=true",
+            self.endpoint,
+            percent_encode_path(&p),
+        );
+        if let Some(auth) = &self.auth {
+            url += &format!("&{auth}");
         }
 
+        let req = Request::post(url)
+            .body(Buffer::new())
+            .map_err(new_request_build_error)?;
+
+        let resp = self.client.send(req).await?;
+
+        let status = resp.status();
+
+        match status {
+            StatusCode::OK => {
+                let bs = resp.into_body();
+                let resp: LocationResponse =
+                    serde_json::from_reader(bs.reader()).map_err(new_json_deserialize_error)?;
+
+                Ok(resp.location)
+            }
+            _ => Err(parse_error(resp)),
+        }
+    }
+
+    pub async fn webhdfs_rename_object(&self, from: &str, to: &str) -> Result<Response<Buffer>> {
+        let from = build_abs_path(&self.root, from);
+        let to = build_rooted_abs_path(&self.root, to);
+
+        let mut url = format!(
+            "{}/webhdfs/v1/{}?op=RENAME&destination={}",
+            self.endpoint,
+            percent_encode_path(&from),
+            percent_encode_path(&to)
+        );
+
+        if let Some(auth) = &self.auth {
+            url += &format!("&{auth}");
+        }
+
+        let req = Request::put(&url)
+            .body(Buffer::new())
+            .map_err(new_request_build_error)?;
+
+        self.client.send(req).await
+    }
+
+    pub fn webhdfs_append_request(
+        &self,
+        location: &str,
+        size: u64,
+        body: Buffer,
+    ) -> Result<Request<Buffer>> {
+        let mut url = location.to_string();
+
+        if let Some(auth) = &self.auth {
+            url += &format!("&{auth}");
+        }
+
+        let mut req = Request::post(&url);
+
+        req = req.header(CONTENT_LENGTH, size.to_string());
+
         req.body(body).map_err(new_request_build_error)
     }
 
-    async fn webhdfs_open_request(
+    /// CONCAT will concat sources to the path
+    pub fn webhdfs_concat_request(
         &self,
         path: &str,
-        range: &BytesRange,
-    ) -> Result<Request<AsyncBody>> {
+        sources: Vec<String>,
+    ) -> Result<Request<Buffer>> {
+        let p = build_abs_path(&self.root, path);
+
+        let sources = sources
+            .iter()
+            .map(|p| build_rooted_abs_path(&self.root, p))
+            .collect::<Vec<String>>()
+            .join(",");
+
+        let mut url = format!(
+            "{}/webhdfs/v1/{}?op=CONCAT&sources={}",
+            self.endpoint,
+            percent_encode_path(&p),
+            percent_encode_path(&sources),
+        );
+
+        if let Some(auth) = &self.auth {
+            url += &format!("&{auth}");
+        }
+
+        let req = Request::post(url);
+
+        req.body(Buffer::new()).map_err(new_request_build_error)
+    }
+
+    fn webhdfs_open_request(&self, path: &str, range: &BytesRange) -> Result<Request<Buffer>> {
         let p = build_abs_path(&self.root, path);
         let mut url = format!(
             "{}/webhdfs/v1/{}?op=OPEN",
@@ -252,33 +384,20 @@ impl WebhdfsBackend {
         }
 
         if !range.is_full() {
-            // Webhdfs does not support read from end
-            if range.offset().is_none() && range.size().is_some() {
-                return Err(Error::new(
-                    ErrorKind::Unsupported,
-                    "webhdfs doesn't support read with suffix range",
-                ));
-            };
-
-            if let Some(offset) = range.offset() {
-                url += &format!("&offset={offset}");
-            }
+            url += &format!("&offset={}", range.offset());
             if let Some(size) = range.size() {
                 url += &format!("&length={size}")
             }
         }
 
         let req = Request::get(&url)
-            .body(AsyncBody::Empty)
+            .body(Buffer::new())
             .map_err(new_request_build_error)?;
 
         Ok(req)
     }
 
-    pub async fn webhdfs_list_status_request(
-        &self,
-        path: &str,
-    ) -> Result<Response<IncomingAsyncBody>> {
+    pub async fn webhdfs_list_status_request(&self, path: &str) -> Result<Response<Buffer>> {
         let p = build_abs_path(&self.root, path);
         let mut url = format!(
             "{}/webhdfs/v1/{}?op=LISTSTATUS",
@@ -290,7 +409,7 @@ impl WebhdfsBackend {
         }
 
         let req = Request::get(&url)
-            .body(AsyncBody::Empty)
+            .body(Buffer::new())
             .map_err(new_request_build_error)?;
         self.client.send(req).await
     }
@@ -299,7 +418,7 @@ impl WebhdfsBackend {
         &self,
         path: &str,
         start_after: &str,
-    ) -> Result<Response<IncomingAsyncBody>> {
+    ) -> Result<Response<Buffer>> {
         let p = build_abs_path(&self.root, path);
 
         let mut url = format!(
@@ -315,21 +434,21 @@ impl WebhdfsBackend {
         }
 
         let req = Request::get(&url)
-            .body(AsyncBody::Empty)
+            .body(Buffer::new())
             .map_err(new_request_build_error)?;
         self.client.send(req).await
     }
 
-    async fn webhdfs_read_file(
+    pub async fn webhdfs_read_file(
         &self,
         path: &str,
         range: BytesRange,
-    ) -> Result<Response<IncomingAsyncBody>> {
-        let req = self.webhdfs_open_request(path, &range).await?;
-        self.client.send(req).await
+    ) -> Result<Response<HttpBody>> {
+        let req = self.webhdfs_open_request(path, &range)?;
+        self.client.fetch(req).await
     }
 
-    async fn webhdfs_get_file_status(&self, path: &str) -> Result<Response<IncomingAsyncBody>> {
+    pub(super) async fn webhdfs_get_file_status(&self, path: &str) -> Result<Response<Buffer>> {
         let p = build_abs_path(&self.root, path);
         let mut url = format!(
             "{}/webhdfs/v1/{}?op=GETFILESTATUS",
@@ -342,13 +461,13 @@ impl WebhdfsBackend {
         }
 
         let req = Request::get(&url)
-            .body(AsyncBody::Empty)
+            .body(Buffer::new())
             .map_err(new_request_build_error)?;
 
         self.client.send(req).await
     }
 
-    async fn webhdfs_delete(&self, path: &str) -> Result<Response<IncomingAsyncBody>> {
+    pub async fn webhdfs_delete(&self, path: &str) -> Result<Response<Buffer>> {
         let p = build_abs_path(&self.root, path);
         let mut url = format!(
             "{}/webhdfs/v1/{}?op=DELETE&recursive=false",
@@ -360,7 +479,7 @@ impl WebhdfsBackend {
         }
 
         let req = Request::delete(&url)
-            .body(AsyncBody::Empty)
+            .body(Buffer::new())
             .map_err(new_request_build_error)?;
 
         self.client.send(req).await
@@ -370,9 +489,9 @@ impl WebhdfsBackend {
         let resp = self.webhdfs_get_file_status("/").await?;
         match resp.status() {
             StatusCode::OK => {
-                let bs = resp.into_body().bytes().await?;
+                let bs = resp.into_body();
 
-                let file_status = serde_json::from_slice::<FileStatusWrapper>(&bs)
+                let file_status = serde_json::from_reader::<_, FileStatusWrapper>(bs.reader())
                     .map_err(new_json_deserialize_error)?
                     .file_status;
 
@@ -386,22 +505,23 @@ impl WebhdfsBackend {
             StatusCode::NOT_FOUND => {
                 self.create_dir("/", OpCreateDir::new()).await?;
             }
-            _ => return Err(parse_error(resp).await?),
+            _ => return Err(parse_error(resp)),
         }
         Ok(())
     }
 }
 
-#[async_trait]
-impl Accessor for WebhdfsBackend {
-    type Reader = IncomingAsyncBody;
-    type BlockingReader = ();
-    type Writer = oio::OneShotWriter<WebhdfsWriter>;
-    type BlockingWriter = ();
+impl Access for WebhdfsBackend {
+    type Reader = HttpBody;
+    type Writer = WebhdfsWriters;
     type Lister = oio::PageLister<WebhdfsLister>;
+    type Deleter = oio::OneShotDeleter<WebhdfsDeleter>;
+    type BlockingReader = ();
+    type BlockingWriter = ();
     type BlockingLister = ();
+    type BlockingDeleter = ();
 
-    fn info(&self) -> AccessorInfo {
+    fn info(&self) -> Arc<AccessorInfo> {
         let mut am = AccessorInfo::default();
         am.set_scheme(Scheme::Webhdfs)
             .set_root(&self.root)
@@ -409,28 +529,26 @@ impl Accessor for WebhdfsBackend {
                 stat: true,
 
                 read: true,
-                read_can_next: true,
-                read_with_range: true,
 
                 write: true,
+                write_can_append: true,
+                write_can_multi: self.atomic_write_dir.is_some(),
+
                 create_dir: true,
                 delete: true,
 
                 list: true,
 
+                shared: true,
+
                 ..Default::default()
             });
-        am
+        am.into()
     }
 
     /// Create a file or directory
     async fn create_dir(&self, path: &str, _: OpCreateDir) -> Result<RpCreateDir> {
-        let req = self.webhdfs_create_object_request(
-            path,
-            Some(0),
-            &OpWrite::default(),
-            AsyncBody::Empty,
-        )?;
+        let req = self.webhdfs_create_dir_request(path)?;
 
         let resp = self.client.send(req).await?;
 
@@ -442,9 +560,9 @@ impl Accessor for WebhdfsBackend {
         // the redirection should be done automatically.
         match status {
             StatusCode::CREATED | StatusCode::OK => {
-                let bs = resp.into_body().bytes().await?;
+                let bs = resp.into_body();
 
-                let resp = serde_json::from_slice::<BooleanResp>(&bs)
+                let resp = serde_json::from_reader::<_, BooleanResp>(bs.reader())
                     .map_err(new_json_deserialize_error)?;
 
                 if resp.boolean {
@@ -456,48 +574,10 @@ impl Accessor for WebhdfsBackend {
                     ))
                 }
             }
-            _ => Err(parse_error(resp).await?),
+            _ => Err(parse_error(resp)),
         }
     }
 
-    async fn read(&self, path: &str, args: OpRead) -> Result<(RpRead, Self::Reader)> {
-        let range = args.range();
-        let resp = self.webhdfs_read_file(path, range).await?;
-        match resp.status() {
-            StatusCode::OK | StatusCode::PARTIAL_CONTENT => {
-                let size = parse_content_length(resp.headers())?;
-                let range = parse_content_range(resp.headers())?;
-                Ok((
-                    RpRead::new().with_size(size).with_range(range),
-                    resp.into_body(),
-                ))
-            }
-            // WebHDFS will returns 403 when range is outside of the end.
-            StatusCode::FORBIDDEN => {
-                let (parts, body) = resp.into_parts();
-                let bs = body.bytes().await?;
-                let s = String::from_utf8_lossy(&bs);
-                if s.contains("out of the range") {
-                    Ok((RpRead::new(), IncomingAsyncBody::empty()))
-                } else {
-                    Err(parse_error_msg(parts, &s)?)
-                }
-            }
-            StatusCode::RANGE_NOT_SATISFIABLE => {
-                resp.into_body().consume().await?;
-                Ok((RpRead::new().with_size(Some(0)), IncomingAsyncBody::empty()))
-            }
-            _ => Err(parse_error(resp).await?),
-        }
-    }
-
-    async fn write(&self, path: &str, args: OpWrite) -> Result<(RpWrite, Self::Writer)> {
-        Ok((
-            RpWrite::default(),
-            oio::OneShotWriter::new(WebhdfsWriter::new(self.clone(), args, path.to_string())),
-        ))
-    }
-
     async fn stat(&self, path: &str, _: OpStat) -> Result<RpStat> {
         // if root exists and is a directory, stat will be ok
         self.root_checker
@@ -508,9 +588,9 @@ impl Accessor for WebhdfsBackend {
         let status = resp.status();
         match status {
             StatusCode::OK => {
-                let bs = resp.into_body().bytes().await?;
+                let bs = resp.into_body();
 
-                let file_status = serde_json::from_slice::<FileStatusWrapper>(&bs)
+                let file_status = serde_json::from_reader::<_, FileStatusWrapper>(bs.reader())
                     .map_err(new_json_deserialize_error)?
                     .file_status;
 
@@ -526,22 +606,50 @@ impl Accessor for WebhdfsBackend {
                 Ok(RpStat::new(meta))
             }
 
-            _ => Err(parse_error(resp).await?),
+            _ => Err(parse_error(resp)),
         }
     }
 
-    async fn delete(&self, path: &str, _: OpDelete) -> Result<RpDelete> {
-        let resp = self.webhdfs_delete(path).await?;
+    async fn read(&self, path: &str, args: OpRead) -> Result<(RpRead, Self::Reader)> {
+        let resp = self.webhdfs_read_file(path, args.range()).await?;
 
-        match resp.status() {
-            StatusCode::OK => {
-                resp.into_body().consume().await?;
-                Ok(RpDelete::default())
+        let status = resp.status();
+
+        match status {
+            StatusCode::OK | StatusCode::PARTIAL_CONTENT => {
+                Ok((RpRead::default(), resp.into_body()))
+            }
+            _ => {
+                let (part, mut body) = resp.into_parts();
+                let buf = body.to_buffer().await?;
+                Err(parse_error(Response::from_parts(part, buf)))
             }
-            _ => Err(parse_error(resp).await?),
         }
     }
 
+    async fn write(&self, path: &str, args: OpWrite) -> Result<(RpWrite, Self::Writer)> {
+        let w = WebhdfsWriter::new(self.clone(), args.clone(), path.to_string());
+
+        let w = if args.append() {
+            WebhdfsWriters::Two(oio::AppendWriter::new(w))
+        } else {
+            WebhdfsWriters::One(oio::BlockWriter::new(
+                w,
+                args.executor().cloned(),
+                args.concurrent(),
+            ))
+        };
+
+        Ok((RpWrite::default(), w))
+    }
+
+    async fn delete(&self) -> Result<(RpDelete, Self::Deleter)> {
+        Ok((
+            RpDelete::default(),
+            oio::OneShotDeleter::new(WebhdfsDeleter::new(Arc::new(self.clone()))),
+        ))
+    }
+
     async fn list(&self, path: &str, args: OpList) -> Result<(RpList, Self::Lister)> {
         if args.recursive() {
             return Err(Error::new(
@@ -555,3 +663,9 @@ impl Accessor for WebhdfsBackend {
         Ok((RpList::default(), oio::PageLister::new(l)))
     }
 }
+
+#[derive(Debug, Deserialize)]
+#[serde(rename_all = "PascalCase")]
+pub(super) struct LocationResponse {
+    pub location: String,
+}
